#include "st_i.h"
#include "g72x.h"
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#include <time.h>

#define RING_ENCODING_ULAW 0
#define RING_ENCODING_G726 2
#define RING_ENCODING_GSM  3
#define RING_ENCODING_G723 4
#define RING_ENCODING_ALAW 8
#define RING_ENCODING_G722 9
#define RING_ENCODING_G728 15
#define RING_ENCODING_ILBC 98

#define RING_MAX_SIZE 65536
#define RING_HEADER_SIZE 512

/* Private data */
typedef struct ringpriv
{
    /* For writer: size in bytes */
    st_size_t data_size;
    /* Checksum */
    uint16_t checksum;
    /* Version */
    st_size_t version;
    /* Year */
    uint16_t year;
    /* Month */
    uint8_t month;
    /* Day */
    uint8_t day;
    /* Hour */
    uint8_t hour;
    /* Minute */
    uint8_t minute;
    /* Filename */
    char filename[16];
    /* Codec */
    uint16_t codec;
    /* Padding */
    uint8_t padding[478];
    /* For seeking */
    st_size_t dataStart;
    /* For G72x decoding */
    struct g72x_state state;
    int ( *dec_routine )();
    int dec_bits;
    unsigned int in_buffer;
    int in_bits;
} *ring_t;

static void ringwriteheader( ft_t ft );
static void ringwriteheadersizeandchecksum( ft_t ft, st_size_t data_size );

static int st_ringencodingsetting( uint16_t ring_encoding, char* encoding, char* size, st_rate_t* rate, char* channels )
{
    switch ( ring_encoding )
    {
        case RING_ENCODING_ULAW:
            *encoding = ST_ENCODING_ULAW;
            *size = ST_SIZE_BYTE;
            *channels = 1;
            *rate = 8000;
            break;

        case RING_ENCODING_ALAW:
            *encoding = ST_ENCODING_ALAW;
            *size = ST_SIZE_BYTE;
            *channels = 1;
            *rate = 8000;
            break;

        default:
            st_report( "encoding: 0x%lx", ring_encoding );
            return ST_EOF;
    }

    return ST_SUCCESS;
}

int st_ringseek( ft_t ft, st_size_t offset )
{
    ring_t ring = ( ring_t )ft->priv;

    if ( ring->dec_routine != NULL )
        st_fail_errno( ft, ST_ENOTSUP, "Sorry, DEC unsupported" );
    else
        return st_seek( ft, offset * ft->info.size + RING_HEADER_SIZE, SEEK_SET );

    return ft->st_errno;
}

int st_ringstartread( ft_t ft ) 
{
    ring_t p = ( ring_t )ft->priv;
    int rc;

    /* 
     * RING is in big endian format.  Swap whats read
     * in on little endian machines.
     */
    if ( ST_IS_LITTLEENDIAN )
        ft->swap = ft->swap ? 0 : 1;

    /* Read the ring file size */
    st_readdw( ft, &p->data_size );
    p->data_size = p->data_size * sizeof( uint16_t ) - RING_HEADER_SIZE;

    /* Read the checksum */
    st_readw( ft, &p->checksum );

    /* Read the version */
    st_readdw( ft, &p->version );

    /* Read the time stamp */
    st_readw( ft, &p->year );
    st_readb( ft, &p->month );
    st_readb( ft, &p->day );
    st_readb( ft, &p->hour );
    st_readb( ft, &p->minute );

    /* Read the file name */
    st_read( ft, p->filename, sizeof( char ), 16 );
    if ( strcmp( p->filename, "ring.bin" ) )
    {
        st_fail_errno( ft, ST_EHDR, "Invalid filename in Grandstream ring tone header" );
        return ST_EOF;
    }

    /* Read the codec */
    st_readw( ft, &p->codec );

    /* Read the padding */
    st_read( ft, p->padding, sizeof( uint8_t ), sizeof( p->padding ) );

    /* Translate the encoding into encoding and size parameters */
    /* (Or, for G.72x, set the decoding routine and parameters) */
    p->dec_routine = NULL;
    p->in_buffer = 0;
    p->in_bits = 0;
    if ( st_ringencodingsetting( p->codec, &( ft->info.encoding ), &( ft->info.size ), &( ft->info.rate ), &( ft->info.channels ) ) == ST_EOF )
    {
        st_fail_errno( ft, ST_EFMT, "Unsupported encoding in Grandstream ring tone header." );
        return ST_EOF;
    }

    switch ( p->codec )
    {
        default:
            break;
    }

    /* Needed for seeking */
    ft->length = p->data_size / ft->info.size;
    if ( ft->seekable )
        p->dataStart = ftell( ft->fp );

    /* Needed for rawread() */
    rc = st_rawstartread( ft );
    if ( rc )
        return rc;

    return ST_SUCCESS;
}

/* 
   When writing, the header is supposed to contain the number of
   data bytes written, unless it is written to a pipe.
   Since we don't know how many bytes will follow until we're done,
   we first write the header with an unspecified number of bytes,
   and at the end we rewind the file and write the header again
   with the right size.  This only works if the file is seekable;
   if it is not, the unspecified size remains in the header
   (this is legal).
*/
int st_ringstartwrite( ft_t ft ) 
{
    ring_t p = ( ring_t )ft->priv;
    int rc;
    int i;

    /* Needed because of rawwrite(); */
    rc = st_rawstartwrite( ft );
    if ( rc )
        return rc;

    /* 
     * Ring file is in big endian format.  Swap whats read in
     * on little endian machines.
     */
    if ( ST_IS_LITTLEENDIAN )
        ft->swap = ft->swap ? 0 : 1;

    p->data_size = 0;
    p->checksum = 0;
    for ( i = 0; i < sizeof( p->padding ); i++ )
        p->padding[i] = 0;
    ringwriteheader( ft );
    return ST_SUCCESS;
}

/*
 * Unpack input codes and pass them back as bytes.
 * Returns 1 if there is residual input, returns -1 if eof, else returns 0.
 * (Adapted from Sun's decode.c.)
 */
static int unpack_input( ft_t ft, unsigned char* code )
{
    ring_t p = ( ring_t )ft->priv;
    unsigned char in_byte;

    if ( p->in_bits < p->dec_bits )
    {
        if ( st_readb( ft, &in_byte ) == ST_EOF )
        {
            *code = 0;
            return -1;
        }
        p->in_buffer |= ( in_byte << p->in_bits );
        p->in_bits += 8;
    }
    *code = p->in_buffer & ( ( 1 << p->dec_bits ) - 1 );
    p->in_buffer >>= p->dec_bits;
    p->in_bits -= p->dec_bits;
    return ( p->in_bits > 0 );
}

st_ssize_t st_ringread( ft_t ft, st_sample_t* buf, st_ssize_t samp )
{
    ring_t p = ( ring_t )ft->priv;
    unsigned char code;
    int done = 0;

    if ( p->dec_routine == NULL )
        return st_rawread( ft, buf, samp );
    while ( samp > 0 && unpack_input( ft, &code ) >= 0 )
    {
        *buf++ = ST_SIGNED_WORD_TO_SAMPLE( ( *p->dec_routine )( code, AUDIO_ENCODING_LINEAR, &p->state ) );
        samp--;
        done++;
    }
    return done;
}

st_ssize_t st_ringwrite( ft_t ft, st_sample_t* buf, st_ssize_t samp )
{
    ring_t p = ( ring_t )ft->priv;
    st_size_t size = p->data_size + samp * ft->info.size;

    if ( p->data_size < ( RING_MAX_SIZE - RING_HEADER_SIZE ) )
    {
        if ( size <= ( RING_MAX_SIZE - RING_HEADER_SIZE ) )
        {
            p->data_size = size;
            return st_rawwrite( ft, buf, samp );
        }
        else
        {
            size = ( RING_MAX_SIZE - RING_HEADER_SIZE ) - p->data_size;
            p->data_size = ( RING_MAX_SIZE - RING_HEADER_SIZE );
            samp = size / ft->info.size;
            return st_rawwrite( ft, buf, samp );
        }
    }
    else
        return 0;
}

int st_ringstopwrite( ft_t ft )
{
    ring_t p = ( ring_t )ft->priv;
    int rc;
    int i;

    /*
     * Needed because of rawwrite(). Do now to flush
     * data before seeking around below.
     */
    rc = st_rawstopwrite( ft );
    if ( rc )
        return rc;

    /* Attempt to update header */
    if ( ft->seekable )
    {
        if ( fseek( ft->fp, RING_HEADER_SIZE, SEEK_SET ) != 0 )
        {
            st_fail_errno( ft, errno, "Can't seek to ring tone data to calculate checksum." );
            return ST_EOF;
        }
        for ( i = 0; i < ( p->data_size / 2 ); i++ )
        {
            uint16_t val;

            st_readw( ft, &val );
            p->checksum += val;
        }
        if ( ( p->data_size % 2 ) > 0 )
        {
            uint8_t val;

            st_readb( ft, &val );
            p->checksum += ( uint16_t )( val << 8 );
        }
        if ( fseek( ft->fp, 0L, SEEK_SET ) != 0 )
        {
            st_fail_errno( ft, errno, "Can't rewind output file to rewrite Grandstream ring tone header." );
            return ST_EOF;
        }
        ringwriteheadersizeandchecksum( ft, p->data_size );
        if ( ( p->data_size % 2 ) > 0 )
        {
            if ( fseek( ft->fp, 0L, SEEK_END ) != 0 )
            {
                st_fail_errno( ft, errno, "Can't rewind output file to rewrite Grandstream ring tone header." );
                return ST_EOF;
            }
            if ( fwrite( p->padding, sizeof( uint8_t ), 1, ft->fp ) != 1 )
            {
                st_fail_errno( ft, errno, "Can't write trailing padding byte to Grandstream ring tone." );
                return ST_EOF;
            }
        }
    }
    return ST_SUCCESS;
}

static int st_ringencoding( int size, int encoding )
{
    int ring_encoding;

    if ( encoding == ST_ENCODING_ULAW && size == ST_SIZE_BYTE )
        ring_encoding = RING_ENCODING_ULAW;
    else if ( encoding == ST_ENCODING_ALAW && size == ST_SIZE_BYTE )
        ring_encoding = RING_ENCODING_ALAW;
    else
        ring_encoding = -1;

    return ring_encoding;
}

static void ringwriteheadersizeandchecksum( ft_t ft, st_size_t data_size )
{
    ring_t p = ( ring_t )ft->priv;
    uint32_t size = ( data_size + RING_HEADER_SIZE + 1 ) / 2;

    st_writedw( ft, size );
    p->checksum += ( size >> 16 ) & 0x00ffff;
    p->checksum += size & 0x00ffff;
    p->checksum = ( uint32_t )0x010000 - ( uint32_t )p->checksum;
    st_writew( ft, p->checksum );
}

static void ringwriteheader( ft_t ft )
{
    ring_t p = ( ring_t )ft->priv;
    time_t now = time( NULL );
    struct tm* dttm = localtime( &now );
    int encoding;
    int i;

    if ( ( encoding = st_ringencoding( ft->info.size, ft->info.encoding ) ) == -1 )
    {
        st_report( "Unsupported output encoding/size for Grandstream ring tone header." );
        st_report( "Defaulting to 8khz u-law\n" );
        encoding = RING_ENCODING_ULAW;
    }
    p->codec = encoding;
    switch ( encoding )
    {
        case RING_ENCODING_ULAW:
            ft->info.encoding = ST_ENCODING_ULAW;
            ft->info.size = ST_SIZE_BYTE;
            ft->info.channels = 1;
            ft->info.rate = 8000;
            break;

        case RING_ENCODING_ALAW:
            ft->info.encoding = ST_ENCODING_ALAW;
            ft->info.size = ST_SIZE_BYTE;
            ft->info.channels = 1;
            ft->info.rate = 8000;
            break;
    }

    st_writedw( ft, 0 );
    st_writew( ft, 0 );
    st_writedw( ft, 0x01000000 );
    p->checksum += 0x0100;
    st_writew( ft, ( uint16_t )( 1900 + dttm->tm_year ) );
    p->checksum += 1900 + dttm->tm_year;
    st_writeb( ft, ( uint8_t )( dttm->tm_mon + 1 ) );
    p->checksum += ( uint16_t )( ( dttm->tm_mon + 1 ) << 8 );
    st_writeb( ft, ( uint8_t )dttm->tm_mday );
    p->checksum += ( uint16_t )dttm->tm_mday;
    st_writeb( ft, ( uint8_t )dttm->tm_hour );
    p->checksum += ( uint16_t )( dttm->tm_hour << 8 );
    st_writeb( ft, ( uint8_t )dttm->tm_min );
    p->checksum += ( uint16_t )dttm->tm_min;
    st_writes( ft, "ring.bin" );
    p->checksum += 0x7269;
    p->checksum += 0x6e67;
    p->checksum += 0x2e62;
    p->checksum += 0x696e;
    st_writeb( ft, 0 );
    st_writeb( ft, 0 );
    st_writeb( ft, 0 );
    st_writeb( ft, 0 );
    st_writeb( ft, 0 );
    st_writeb( ft, 0 );
    st_writeb( ft, 0 );
    st_writeb( ft, 0 );
    st_writew( ft, p->codec );
    p->checksum += ( uint16_t )p->codec;
    for ( i = 0; i < 239; i++ )
        st_writew( ft, 0 );
}

